@@ -1,6 +1,6 @@
 # python-musicpd: Python MPD client library
 # Copyright (C) 2008-2010  J. Alexander Treuman <jat@spatialrift.net>
-# Copyright (C) 2012-2013  Kaliko Jack <kaliko@azylum.org>
+# Copyright (C) 2012-2018  Kaliko Jack <kaliko@azylum.org>
 #
 # python-musicpd is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -15,16 +15,37 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with python-musicpd.  If not, see <http://www.gnu.org/licenses/>.
 
-# pylint: disable=C0111
+# pylint: disable=missing-docstring
 
 import socket
+import os
+
+from functools import wraps
 
 
 HELLO_PREFIX = "OK MPD "
 ERROR_PREFIX = "ACK "
 SUCCESS = "OK"
 NEXT = "list_OK"
-VERSION = '0.4.1'
+VERSION = '0.4.3'
+
+
+def iterator_wrapper(func):
+    """Decorator handling iterate option"""
+    @wraps(func)
+    def decorated_function(instance, *args, **kwargs):
+        generator = func(instance, *args, **kwargs)
+        if not instance.iterate:
+            return list(generator)
+        instance._iterating = True
+        def iterator(gen):
+            try:
+                for item in gen:
+                    yield item
+            finally:
+                instance._iterating = False
+        return iterator(generator)
+    return decorated_function
 
 
 class MPDError(Exception):
@@ -55,6 +76,8 @@ class Range:
         self._check()
 
     def __str__(self):
+        if len(self.tpl) == 0:
+            return ':'
         if len(self.tpl) == 1:
             return '{0}:'.format(self.tpl[0])
         return '{0[0]}:{0[1]}'.format(self.tpl)
@@ -65,15 +88,14 @@ class Range:
     def _check(self):
         if not isinstance(self.tpl, tuple):
             raise CommandError('Wrong type, provide a tuple')
-        if len(self.tpl) not in [1, 2]:
-            raise CommandError('length not in [1, 2]')
+        if len(self.tpl) not in [0, 1, 2]:
+            raise CommandError('length not in [0, 1, 2]')
         for index in self.tpl:
             try:
                 index = int(index)
             except (TypeError, ValueError):
                 raise CommandError('Not a tuple of int')
 
-
 class _NotConnected:
     def __getattr__(self, attr):
         return self._dummy
@@ -82,6 +104,24 @@ class _NotConnected:
         raise ConnectionError("Not connected")
 
 class MPDClient:
+    """MPDClient instance will look for ``MPD_HOST``/``MPD_PORT``/``XDG_RUNTIME_DIR`` environment
+    variables and set instance attribute ``host``, ``port`` and ``pwd``
+    accordingly.
+
+    Then :py:obj:`musicpd.MPDClient.connect` will use ``host`` and ``port`` as defaults if not provided as args.
+
+    Cf. :py:obj:`musicpd.MPDClient.connect` for details.
+
+    >>> from os inport environ
+    >>> environ['MPD_HOST'] = 'pass@mpdhost'
+    >>> cli = musicpd.MPDClient()
+    >>> cli.pwd == environ['MPD_HOST'].split('@')[0]
+    True
+    >>> cli.host == environ['MPD_HOST'].split('@')[1]
+    True
+    >>> # cli.connect() will use host/port as set in MPD_HOST/MPD_PORT
+    """
+
     def __init__(self):
         self.iterate = False
         self._reset()
@@ -130,9 +170,14 @@ class MPDClient:
             "playlistsearch":     self._fetch_songs,
             "plchanges":          self._fetch_songs,
             "plchangesposid":     self._fetch_changes,
+            "prio":               self._fetch_nothing,
+            "prioid":             self._fetch_nothing,
+            "rangeid":            self._fetch_nothing,
             "shuffle":            self._fetch_nothing,
             "swap":               self._fetch_nothing,
             "swapid":             self._fetch_nothing,
+            "addtagid":           self._fetch_nothing,
+            "cleartagid":         self._fetch_nothing,
             # Stored Playlist Commands
             "listplaylist":       self._fetch_list,
             "listplaylistinfo":   self._fetch_songs,
@@ -159,6 +204,11 @@ class MPDClient:
             "update":             self._fetch_item,
             "rescan":             self._fetch_item,
             "readcomments":       self._fetch_object,
+            # Mounts and neighbors
+            "mount":              self._fetch_nothing,
+            "unmount":            self._fetch_nothing,
+            "listmounts":         self._fetch_mounts,
+            "listneighbors":      self._fetch_neighbors,
             # Sticker Commands
             "sticker get":        self._fetch_item,
             "sticker set":        self._fetch_nothing,
@@ -176,6 +226,7 @@ class MPDClient:
             "toggleoutput":       self._fetch_nothing,
             "outputs":            self._fetch_outputs,
             # Reflection Commands
+            "config":             self._fetch_object,
             "commands":           self._fetch_list,
             "notcommands":        self._fetch_list,
             "tagtypes":           self._fetch_list,
@@ -188,8 +239,36 @@ class MPDClient:
             "readmessages":       self._fetch_messages,
             "sendmessage":        self._fetch_nothing,
         }
+        self._get_envvars()
+
+    def _get_envvars(self):
+        """
+        Retrieve MPD env. var. to overrides "localhost:6600"
+            Use MPD_HOST/MPD_PORT if set
+            else use MPD_HOST=${XDG_RUNTIME_DIR:-/run/}/mpd/socket if file exists
+        """
+        self.host = 'localhost'
+        self.pwd = None
+        self.port = os.environ.get('MPD_PORT', '6600')
+        mpd_host_env = os.environ.get('MPD_HOST')
+        if mpd_host_env:
+            # If password is set:
+            # mpd_host_env = ['pass', 'host'] because MPD_HOST=pass@host
+            mpd_host_env = mpd_host_env.split('@')
+            mpd_host_env.reverse()
+            self.host = mpd_host_env[0]
+            if len(mpd_host_env) > 1 and mpd_host_env[1]:
+                self.pwd = mpd_host_env[1]
+        else:
+            # Is socket there
+            xdg_runtime_dir = os.environ.get('XDG_RUNTIME_DIR', '/run')
+            rundir = os.path.join(xdg_runtime_dir, 'mpd/socket')
+            if os.path.exists(rundir):
+                self.host = rundir
 
     def __getattr__(self, attr):
+        if attr == 'send_noidle':  # have send_noidle to cancel idle as well as noidle
+            return self.noidle()
         if attr.startswith("send_"):
             command = attr.replace("send_", "", 1)
             wrapper = self._send
@@ -343,19 +422,6 @@ class MPDClient:
             self._command_list = None
         self._fetch_nothing()
 
-    def _iterator_wrapper(self, iterator):
-        try:
-            for item in iterator:
-                yield item
-        finally:
-            self._iterating = False
-
-    def _wrap_iterator(self, iterator):
-        if not self.iterate:
-            return list(iterator)
-        self._iterating = True
-        return self._iterator_wrapper(iterator)
-
     def _fetch_nothing(self):
         line = self._read_line()
         if line is not None:
@@ -367,11 +433,13 @@ class MPDClient:
             return
         return pairs[0][1]
 
+    @iterator_wrapper
     def _fetch_list(self):
-        return self._wrap_iterator(self._read_list())
+        return self._read_list()
 
+    @iterator_wrapper
     def _fetch_playlist(self):
-        return self._wrap_iterator(self._read_playlist())
+        return self._read_playlist()
 
     def _fetch_object(self):
         objs = list(self._read_objects())
@@ -379,8 +447,9 @@ class MPDClient:
             return {}
         return objs[0]
 
+    @iterator_wrapper
     def _fetch_objects(self, delimiters):
-        return self._wrap_iterator(self._read_objects(delimiters))
+        return self._read_objects(delimiters)
 
     def _fetch_changes(self):
         return self._fetch_objects(["cpos"])
@@ -403,8 +472,15 @@ class MPDClient:
     def _fetch_messages(self):
         return self._fetch_objects(["channel"])
 
+    def _fetch_mounts(self):
+        return self._fetch_objects(["mount"])
+
+    def _fetch_neighbors(self):
+        return self._fetch_objects(["neighbor"])
+
+    @iterator_wrapper
     def _fetch_command_list(self):
-        return self._wrap_iterator(self._read_command_list())
+        return self._read_command_list()
 
     def _hello(self):
         line = self._rfile.readline()
@@ -416,7 +492,6 @@ class MPDClient:
         self.mpd_version = line[len(HELLO_PREFIX):].strip()
 
     def _reset(self):
-        # pylint: disable=w0201
         self.mpd_version = None
         self._iterating = False
         self._pending = []
@@ -465,7 +540,29 @@ class MPDClient:
         self._write_command("noidle")
         return self._fetch_list()
 
-    def connect(self, host, port):
+    def connect(self, host=None, port=None):
+        """Connects the MPD server
+
+        :param str host: hostname, IP or FQDN (defaults to `localhost` or socket, see below for details)
+        :param str port: port number (defaults to 6600)
+
+        The connect method honors MPD_HOST/MPD_PORT environment variables.
+
+        .. note:: Default host/port
+
+          If host evaluate to :py:obj:`False`
+           * use ``MPD_HOST`` env. var. if set, extract password if present,
+           * else looks for a existing file in ``${XDG_RUNTIME_DIR:-/run/}/mpd/socket``
+           * else set host to ``localhost``
+
+          If port evaluate to :py:obj:`False`
+           * if ``MPD_PORT`` env. var. is set, use it for port
+           * else use ``6600``
+        """
+        if not host:
+            host = self.host
+        if not port:
+            port = self.port
         if self._sock is not None:
             raise ConnectionError("Already connected")
         if host.startswith("/"):
@@ -481,6 +578,10 @@ class MPDClient:
             raise
 
     def disconnect(self):
+        """Closes the MPD connection.
+        The client closes the actual socket, it does not use the
+        'close' request from MPD protocol (as suggested in documentation).
+        """
         if hasattr(self._rfile, 'close'):
             self._rfile.close()
         if hasattr(self._wfile, 'close'):
@@ -517,5 +618,4 @@ class MPDClient:
 def escape(text):
     return text.replace("\\", "\\\\").replace('"', '\\"')
 
-
 # vim: set expandtab shiftwidth=4 softtabstop=4 textwidth=79:
